# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from qiskit.providers import JobError, JobTimeoutError
from qiskit.providers import JobV1 as Job
from qiskit.providers.jobstatus import JobStatus
from qiskit.result import Result

from ...verifier import BasicTrapInfo, BasicTrapper
from ..base import job_remove, job_result, job_status


class DQPUJob(Job):
    def __init__(self, backend, job_id, options, circuit):
        super().__init__(backend, job_id)
        self._backend = backend
        self.job_id = job_id
        self.options = options
        self.circuit = circuit
        self._results = None

    def submit(self):
        pass

    def _wait_for_result(self, timeout=None, wait=5):
        start_time = time.time()

        while True:
            elapsed = time.time() - start_time
            if timeout and elapsed >= timeout:
                raise JobTimeoutError("Timed out waiting for result")
            status = self.raw_status()
            if status == "executed":
                break
            if status == "invalid":
                raise JobError("Job has been marked as invalid")
            time.sleep(wait)

        if self._results is None:
            self._results = job_result(
                self._backend.near_blockchain, self._backend.ipfs_gateway, self.job_id
            )
        return self._results

    def result(self, timeout=None, wait=5):
        counts, trap_list = self._wait_for_result(timeout, wait)

        # TODO: handle different trap classes
        traps = list(map(BasicTrapInfo.loads, trap_list))
        counts = BasicTrapper().untrap_results(traps, counts)

        results = [
            {
                "success": True,
                "header": {"name": self.circuit.name},
                "shots": len(counts),
                "data": {"counts": counts},
            }
        ]
        return Result.from_dict(
            {
                "results": results,
                "backend_name": self._backend.configuration().backend_name,
                "backend_version": self._backend.configuration().backend_version,
                "job_id": self.job_id,
                "qobj_id": self.circuit.name,
                "success": True,
            }
        )

    def remove(self):
        return job_remove(self._backend.near_blockchain, self.job_id)

    def raw_status(self):
        return job_status(self._backend.near_blockchain, self.job_id)

    def status(self):
        status = self.raw_status()
        if status == "executed":
            status = JobStatus.DONE
        elif status == "invalid":
            status = JobStatus.ERROR
        else:
            status = JobStatus.RUNNING
        return status


def submit(self):
    raise NotImplementedError



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import os
from typing import Optional

from py_near.account import Account
from py_near.exceptions.exceptions import RpcEmptyResponse

from .blockchain import Blockchain

# export type JobStatus = 'pending-validation' | 'waiting' | 'validating-result'
# | 'executed' | 'invalid';

# class Job:
#     id: string
#     owner_id: AccountId
#     reward_amount: bigint
#     sampler_deposit: bigint
#     status: JobStatus

#     qubits: number
#     depth: number
#     shots: number

#     job_file: string
#     result_file: string

#     verifier_id: AccountId
#     sampler_id: AccountId


def to_near(v):
    return int(v * 1000000000000000000000000)


def from_near(v):
    return int(v) / 1000000000000000000000000.0


def asyncio_run_nested(v, n_repeat=5):
    try:
        return asyncio.run(v())
    except RpcEmptyResponse as e:
        print(e)
        raise RpcEmptyResponse()
    except Exception as e:
        print(e)
        import nest_asyncio

        get_ipython()  # type: ignore # noqa: F821
        nest_asyncio.apply()
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(v())


class NearBlockchain(Blockchain):
    def __init__(self, account: str, network="testnet"):
        self.network = network

        if self.network == "testnet":
            self.contract = "dqpu_7.testnet"
            self.rpc_addr = "https://rpc.testnet.near.org"
        else:
            self.rpc_addr = "https://rpc.mainnet.near.org"

        self.account = self.load_account(account)

    def balance(self):
        async def v():
            return await self.account.get_balance()

        return from_near(asyncio.run(v()))

    def create_account(self, name: str, useFaucet: bool = True):
        pass

    def load_account(self, account: str):
        fn = account
        if "/" not in account:
            fn = os.path.expanduser(
                os.path.join("~", ".near-credentials", self.network, account + ".json")
            )

        if not os.path.exists(fn):
            raise Exception(f"Unable to load wallet file: {fn}")

        with open(fn, "r") as wf:
            w = json.loads(wf.read())

            acc = Account(w["account_id"], w["private_key"], self.rpc_addr)

            async def v():
                await acc.startup()

            asyncio_run_nested(v)

            return acc

    def view(self, view_name: str, params):
        async def v():
            return await self.account.view_function(self.contract, view_name, params)

        return asyncio_run_nested(v).result

    def call(self, function_name: str, params, amount=0):
        async def v():
            return await self.account.function_call(
                self.contract,
                function_name,
                params,
                amount=to_near(amount),
            )

        r = asyncio_run_nested(v)

        if "Failure" in r.status:
            raise Exception(r.status["Failure"])
        return r.transaction.hash

    # Submit a new quantum job
    def submit_job(self, qubits, depth, shots, job_file, reward):
        return self.call(
            "submit_job",
            {"qubits": qubits, "depth": depth, "shots": shots, "job_file": job_file},
            reward,
        )

    # Remove a pending-validation or waiting job
    def remove_job(self, id):
        return self.call("remove_job", {"id": id})

    # Called by validators, set the validity of a pending-validation job
    def set_job_validity(
        self, id: int, valid: bool, trapped_file: Optional[str] = None
    ):
        return self.call(
            "set_job_validity", {"id": id, "valid": valid, "trapped_file": trapped_file}
        )

    # Submit a result for a waiting job, with the caution
    def submit_job_result(self, id: int, result_file: str, deposit: int):
        return self.call(
            "submit_job_result", {"id": id, "result_file": result_file}, deposit
        )

    # Called by validators, set the validity of a job result for a 'validating-result' job
    def set_result_validity(self, id: int, valid: bool, trap_file: str = ""):
        return self.call(
            "set_result_validity", {"id": id, "valid": valid, "trap_file": trap_file}
        )

    # Get latest quantum job list
    def get_latest_jobs(self, limit=50):
        return self.view("get_latest_jobs", {"limit": limit})

    def get_jobs(self, from_index=0, limit=50):
        return self.view("get_jobs", {"from_index": from_index, "limit": limit})

    # Get a single quantum job by its id
    def get_job(self, id: str):
        return self.view("get_job", {"id": id})

    # Get a quantum job status by its id
    def get_job_status(self, id: str):
        return self.view("get_job_status", {"id": id})

    def get_number_of_jobs(self):
        return self.view("get_number_of_jobs", {})

    def get_number_of_verifiers(self):
        return self.view("get_number_of_verifiers", {})

    def get_handled_amount(self):
        return self.view("get_handled_amount", {})

    # Clear all jobs
    def clear_jobs(self):
        return self.call("clear_jobs", {})

    def get_jobs_stats(self, limit=1000):
        return self.view("get_jobs_stats", {"limit": limit})

    # Add a verifier
    def add_verifier(self, account):
        return self.call("add_verifier", {"account": account})

    # Remove a verifier
    def remove_verifier(self, account):
        return self.call("remove_verifier", {"account": account})

    # Return true if the account is a verifier
    def is_a_verifier(self, account):
        return self.view("is_a_verifier", {"account": account})

    def get_verifiers(self):
        return self.view("get_verifiers", {})

    # Change contract owner
    def set_owner(self, account):
        return self.call("set_owner", {"new_owner": account})



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import random
from hashlib import sha256

from .blockchain import (
    IPFSGateway,
    NearBlockchain,
)  # , start_ipfs_daemon, stop_ipfs_daemon
from .q import Circuit
from .utils import create_dqpu_dirs

# from qiskit import qasm2
# from qiskit.circuit.random import random_circuit


ACTIONS = [
    "list",
    "submit",
    "remove",
    "info",
    "status",
    "get-result",
    "set-validity",
    "set-result-validity",
    "submit-result",
    "submit-random",
    "clear-jobs",
    "add-verifier",
    "remove-verifier",
    "is-a-verifier",
    "verifiers",
]


def default_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n", "--network", help="network to use (mainnet | testnet)", default="testnet"
    )
    parser.add_argument(
        "-ig",
        "--ipfs-gateway",
        help="ipfs gateway (default 127.0.0.1:8080)",
        default="127.0.0.1:8080",
    )
    parser.add_argument("-a", "--account", help="account path or name", required=True)
    return parser


def cli():  # noqa: C901
    parser = default_parser()

    parser.add_argument("action", help="action to run", choices=ACTIONS)

    parser.add_argument("-i", "--id", help="identifier of job")
    parser.add_argument("-v", "--validity", help="validity status", type=str)
    parser.add_argument("-vv", "--verifier", help="verifier account", type=str)
    parser.add_argument(
        "-mq",
        "--max-qubits",
        help="max qubits for the random submit",
        type=int,
        default=21,
    )
    parser.add_argument(
        "-nq",
        "--min-qubits",
        help="min qubits for the random submit",
        type=int,
        default=5,
    )

    group_submit = parser.add_argument_group("submit")
    group_submit.add_argument("-f", "--file", help="openqasm2 file to submit")
    group_submit.add_argument("-t", "--trap", help="the trap file", default="")
    group_submit.add_argument(
        "-s", "--shots", help="number of shots to request", type=int, default=1024
    )
    group_submit.add_argument(
        "-r",
        "--reward",
        help="reward for the execution of the job",
        type=float,
        default=0.001,
    )

    # group_remove = parser.add_argument_group("remove")

    # group_job_validity = parser.add_argument_group("set-job-validity")

    group_submit_result = parser.add_argument_group("submit-result")
    group_submit_result.add_argument(
        "-rf", "--result-file", help="json containing sampling result"
    )
    group_submit_result.add_argument(
        "-d",
        "--deposit",
        help="cautional deposit for the execuition of the job",
        type=float,
        default=None,
    )

    args = parser.parse_args()  # noqa: F841
    base_dir = create_dqpu_dirs()

    nb = NearBlockchain(args.account, args.network)
    ipfs = IPFSGateway(gateway="http://" + args.ipfs_gateway)  # noqa: F841

    # ipfs_process = start_ipfs_daemon()

    if args.action == "list":
        print(json.dumps(nb.get_latest_jobs(), indent=2))

    elif args.action == "submit":
        # Parse the qasm
        qf = open(args.file, "r")
        qasm_data = qf.read()
        circ = Circuit.fromQasmCircuit(qasm_data)

        # Calcualte qubits, depth
        qubits = circ.n_qbits
        depth = len(circ.gates)

        # Upload job_file
        job_file = ipfs.upload(args.file)

        print(nb.submit_job(qubits, depth, args.shots, job_file, args.reward))
        print(nb.get_latest_jobs()[0]["id"])

    elif args.action == "submit-random":
        nq = random.randint(int(args.min_qubits), int(args.max_qubits))
        dpt = random.randint(5, 300)

        print(f"Creating a random circuit of {nq} qubits (depth {dpt})...")

        qc = Circuit.random(nq, dpt)
        qasm_data = qc.toQasmCircuit()

        # qc = random_circuit(nq, dpt, max_operands=2, measure=True)
        # qasm_data = qasm2.dumps(qc)

        dig = sha256(qasm_data.encode("ascii")).hexdigest()[0:8]

        circuit_file = base_dir + f"/cache/random_circuit_{nq}_{dpt}_{dig}.qasm"
        with open(circuit_file, "w") as f:
            f.write(qasm_data)

        print("Submitting...")

        # Upload job_file
        job_file = ipfs.upload(circuit_file)
        print(f"Circuit file is {job_file}")

        shots = random.choice([256, 512, 1024, 2048, 4096, 8192, 16384])

        # sh_fact = (shots / 16384) * nq
        # reward = random.randint(int(0.5 * sh_fact), int(1.1 * sh_fact)) / 1000.0
        reward = nq * nq / 600000
        reward = max(reward, 0.0001)

        print(f"Submitting with a reward of {reward:.6f} for {shots} shots")

        print(nb.submit_job(nq, dpt, shots, job_file, reward))
        print(f"Account balance is {nb.balance():0.5f} N")
        print(nb.get_latest_jobs()[-1]["id"])

    elif args.action == "remove":
        print(nb.remove_job(args.id))

    elif args.action == "info":
        print(json.dumps(nb.get_job(args.id), indent=2))

    elif args.action == "status":
        print(json.dumps(nb.get_job_status(args.id), indent=2))

    elif args.action == "get-result":
        print(ipfs.get(nb.get_job(args.id).result_file))

    elif args.action == "set-validity":
        jid_file = nb.get_job(args.id)["job_file"]

        # TODO: add trap

        print(
            nb.set_job_validity(
                args.id, valid=args.validity == "true", trapped_file=jid_file
            )
        )

    elif args.action == "set-result-validity":
        if args.validity == "true":
            trap_file = ipfs.upload(args.trap)
            print(nb.set_result_validity(args.id, valid=True, trap_file=trap_file))
        else:
            print(nb.set_result_validity(args.id, valid=False))

    elif args.action == "submit-result":
        res_file = ipfs.upload(args.result_file)

        if args.deposit is None:
            args.deposit = nb.get_job(args.id).reward_amount / 10

        print(nb.submit_job_result(args.id, result_file=res_file, deposit=args.deposit))

    elif args.action == "clear-jobs":
        print(nb.clear_jobs())

    elif args.action == "add-verifier":
        print(nb.add_verifier(args.verifier))

    elif args.action == "remove-verifier":
        print(nb.remove_verifier(args.verifier))

    elif args.action == "is-a-verifier":
        print(nb.is_a_verifier(args.verifier))

    elif args.action == "verifiers":
        print(nb.get_verifiers())

    # stop_ipfs_daemon(ipfs_process)



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from . import utils  # noqa: F401
from .circuit import Barrier, Circuit, Measure  # noqa: F401
from .gate import Gate  # noqa: F401
from .gates import Gates  # noqa: F401

ExperimentResult = Dict[str, int]



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qiskit.providers import ProviderV1 as Provider

from .dqpubackend import DQPUBackend

# from qiskit.providers.providerutils import filter_backends


class DQPUProvider(Provider):
    def __init__(self):
        super().__init__()
        self.backend_list = [DQPUBackend(network="testnet", provider=self)]

    def backends(self, name=None, **kwargs):
        backends = self.backend_list
        if name:
            backends = [backend for backend in backends if backend.name() == name]
        return backends  # filter_backends(backends, filters=filters, **kwargs)

    def __str__(self):
        return "DQPUProvider"



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import dask.array as da
# import numpy as np
# from dask.distributed import Client  # , progress
# from scipy.linalg import norm

# from ..q import Gate, Measure
# from .sampler import Sampler

# client = Client(
#     processes=False, threads_per_worker=1, n_workers=1, memory_limit="612MB"
# )


# class DaskSimulatorSampler(Sampler):
#     def __init__(self, circuit, dtype=np.csingle):
#         self.circuit = circuit
#         self.hs = None
#         self.measures = []
#         self.dtype = dtype

#     def compute(self):
#         self.hs = da.zeros(((2,) * self.circuit.n_qbits), dtype=self.dtype)
#         self.hs[(0,) * self.circuit.n_qbits] = 1
#         self.measures = []
#         projectors = [
#             da.array([[1, 0], [0, 0]], dtype=self.dtype),
#             da.array([[0, 0], [0, 1]], dtype=self.dtype),
#         ]

#         for x in self.circuit:
#             a, p = x

#             if isinstance(a, Gate) and a.nq == 2:
#                 pa, pb = p
#                 self.hs = da.tensordot(a._tensor, self.hs, ((2, 3), (pa, pb)))
#                 self.hs = da.moveaxis(self.hs, (0, 1), (pa, pb))

#             elif isinstance(a, Gate) and a.nq == 1:
#                 self.hs = da.tensordot(a.matrix, self.hs, (1, p))
#                 self.hs = da.moveaxis(self.hs, 0, p)

#             elif isinstance(a, Measure):

#                 def project(i, j):
#                     projected = da.tensordot(projectors[j], self.hs, (1, i))
#                     return da.moveaxis(projected, 0, i)

#                 pa, pb = p
#                 projected = project(pa, 0)
#                 norm_projected = norm(projected.flatten())
#                 if da.random.random() < norm_projected**2:
#                     self.hs = projected / norm_projected
#                     self.measures.append((pa, pb, 0))
#                 else:
#                     projected = project(p[0], 1)
#                     self.hs = projected / norm(projected)
#                     self.measures.append((pa, pb, 1))

#         return self.hs.flatten()

#     def compute_sep(self):  # noqa: C901
#         def tensor_product(matrix1, matrix2):
#             m, n = matrix1.shape
#             p, q = matrix2.shape
#             result = da.zeros((m * p, n * q), dtype=matrix1.dtype)
#             for i in range(m):
#                 for j in range(n):
#                     result[i * p : (i + 1) * p, j * q : (j + 1) * q] = (
#                         matrix1[i, j] * matrix2
#                     )
#             return result

#         def perform_quantum_gate(state_vector, gate, qbit):
#             matrices = []
#             for index in range(self.circuit.n_qbits):
#                 matrices.append(gate if index == qbit else np.identity(2))

#             mat = matrices[self.circuit.n_qbits - 1]
#             for index in range(self.circuit.n_qbits - 2, -1, -1):
#                 mat = tensor_product(matrices[index], mat)

#             state_vector = da.dot(state_vector, mat)
#             return state_vector

#         def perform_controlled_gate(state_vector, gate, control_qubit, target_qubit):
#             matrices = []
#             for index in range(self.circuit.n_qbits):
#                 if index == control_qubit:
#                     matrices.append(np.identity(2))
#                 elif index == target_qubit:
#                     matrices.append(gate)
#                 else:
#                     matrices.append(np.identity(2))

#             mat = matrices[self.circuit.n_qbits - 1]
#             for index in range(self.circuit.n_qbits - 2, -1, -1):
#                 mat = tensor_product(matrices[index], mat)

#             new_state_vector = da.zeros_like(state_vector)
#             for i in range(len(state_vector)):
#                 for j in range(len(mat)):
#                     new_state_vector[i] += state_vector[j] * mat[j][i]
#             return new_state_vector

#         self.hs = da.zeros((2**self.circuit.n_qbits), dtype=self.dtype)
#         self.hs[0] = 1
#         projectors = [
#             da.array([[1, 0], [0, 0]], dtype=self.dtype),
#             da.array([[0, 0], [0, 1]], dtype=self.dtype),
#         ]

#         for x in self.circuit:
#             a, p = x

#             if isinstance(a, Gate) and a.nq == 2:
#                 pa, pb = p
#                 self.hs = perform_controlled_gate(
#                     self.hs, a.matrix.reshape(4, 4)[2:, 2:], pa, pb
#                 )

#             elif isinstance(a, Gate) and a.nq == 1:
#                 self.hs = perform_quantum_gate(self.hs, a.matrix, p)

#             elif isinstance(a, Measure):

#                 def project(i, j):
#                     projected = da.tensordot(projectors[j], self.hs, (1, i))
#                     return da.moveaxis(projected, 0, i)

#                 pa, pb = p
#                 projected = project(pa, 0)
#                 norm_projected = norm(projected.flatten())
#                 if da.random.random() < norm_projected**2:
#                     self.hs = projected / norm_projected
#                     self.measures.append((pa, pb, 0))
#                 else:
#                     projected = project(p[0], 1)
#                     self.hs = projected / norm(projected)
#                     self.measures.append((pa, pb, 1))

#         return self.hs.flatten().compute()



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import Optional

from qiskit import qasm2
from qiskit.providers import BackendV2 as Backend
from qiskit.providers import Options, convert_to_target
from qiskit.providers.models import BackendConfiguration

from ...blockchain import IPFSGateway, NearBlockchain
from ..base import submit_job
from .dqpujob import DQPUJob

BASIS_GATES = sorted(
    [
        "x",
        "y",
        "z",
        "h",
        "cx",
        "cy",
        "cz",
        "p",
        # "swap",
        # "cswap",
        # "rx",
        # "ry",
        # "rz",
        # "id",
        # "ccx",
        # "u1",
        # "u2",
        # "u3",
        # "rxx",
        # "ryy",
        # "rzz",
        "s",
        "sdg",
        "t",
        "tdg",
        # "sx",
        # "sxdg",
        # "unitary",
        # "ecr",
    ]
)


class DQPUBackend(Backend):
    def __init__(self, network: str = "testnet", provider=None):
        super().__init__()

        self.network = network
        # self.provider = provider
        self.near_blockchain: Optional[NearBlockchain] = None
        self.ipfs_gateway = IPFSGateway()

        # Create Target
        self._target = convert_to_target(self.configuration())

        # Set option validators
        self.options.set_validator("shots", (1, 8192))
        self.options.set_validator("reward", (0.00001, 10.0))

    def load_account(self, account: str):
        """Load near account given its id or file path"""
        self.near_blockchain = NearBlockchain(account, self.network)

    def configuration(self):
        return BackendConfiguration(
            backend_name="dqpu",
            backend_version="0.1",
            n_qubits=1000,
            basis_gates=BASIS_GATES,
            gates=[],
            local=False,
            simulator=True,
            conditional=True,
            open_pulse=False,
            memory=True,
            max_shots=8192,
            coupling_map=None,
            # , supported_instructions=None, dynamic_reprate_enabled=False,
            # rep_delay_range=None, default_rep_delay=None, max_experiments=None,
            # sample_name=None, n_registers=None, register_map=None, configurable=None,
            # credits_required=None, online_date=None, display_name=None, description=None,
            # tags=None, dt=None, dtm=None, processor_type=None, parametric_pulses=None
        )

    @property
    def target(self):
        return self._target

    @property
    def max_circuits(self):
        return 1

    @classmethod
    def _default_options(cls):
        return Options(shots=1024, reward=0.0001)

    def run(self, circuit, **kwargs):
        # serialize circuits submit to backend and create a job
        for kwarg in kwargs:
            if not hasattr(self.options, kwarg):
                warnings.warn(
                    "Option %s is not used by this backend" % kwarg,
                    # warnings.UserWarning,
                    stacklevel=2,
                )
        options = {
            "shots": kwargs.get("shots", self.options.shots),
            "reward": kwargs.get("reward", self.options.reward),
        }

        qasm_data = qasm2.dumps(circuit)

        job_id = submit_job(
            self.near_blockchain,
            self.ipfs_gateway,
            qasm_data,
            circuit.num_qubits,
            circuit.depth(),
            options,
        )
        return DQPUJob(self, job_id, options, circuit)



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pickle
import random
import time
import traceback

from requests.exceptions import ReadTimeout

from .blockchain import IPFSGateway, NearBlockchain, repeat_until_done
from .cli import default_parser
from .q import Circuit
from .utils import create_dqpu_dirs
from .verifier import BasicTrapper  # BasicTrapInfo,


def handle_pending_validation_job(j, ipfs, nb, base_dir):
    print(f"Processing pending-validation job {j['id']} from {j['owner_id']}")

    try:
        jf = ipfs.get(j["job_file"], timeout=10)
    except ReadTimeout:  # TODO: move on ipfs.get raising a new exception
        print(f"\tTimeout getting file {j['job_file']}, skipping for now")
        return False

    # Parse the file using q.Circuit.fromQasm
    try:
        qc = Circuit.fromQasmCircuit(jf.decode("ascii"))
    except Exception as e:
        print("\t", "Failed to parse", j["id"], e)
        nb.set_job_validity(j["id"], False)
        return True

    # Add a trap
    trapper = BasicTrapper()
    (qc2, trap_list) = trapper.trap(qc)

    # Save trap info to a file
    with open(f"{base_dir}/{j['id']}_qc_traps.pickle", "wb") as outp:
        pickle.dump(trap_list, outp, pickle.HIGHEST_PROTOCOL)

    # Save qasm to file
    trapped_qasm_file = f"{base_dir}/{j['id']}_qc_trapped.qasm"

    with open(trapped_qasm_file, "w") as f:
        f.write(qc2.toQasmCircuit())

    # Upload the file
    jf_trapped = ipfs.upload(trapped_qasm_file)
    print("\tTrapped file uploaded", jf_trapped)

    # Send the set_validity
    print("\t", nb.set_job_validity(j["id"], True, jf_trapped))
    return True


def handle_validating_result_job(j, ipfs, nb, base_dir):  # noqa: C901
    print(
        f"Processing validating-result job {j['id']} from {j['owner_id']} "
        + f"sampled by {j['sampler_id']}"
    )

    # Get the result data
    try:
        rf = ipfs.get(j["result_file"], timeout=10)
    except ReadTimeout:  # TODO: move on ipfs.get raising a new exception
        print(f"\tTimeout getting file {j['result_file']}, skipping for now")
        return False

    try:
        counts = json.loads(rf)
    except:
        print("\tInvalid result data")
        print(nb.set_result_validity(j["id"], False))
        return True

    # Check if shots match
    ctot = sum(counts.values())
    if j["shots"] > ctot:
        print("\t", "Invalid number of shots")
        print(nb.set_result_validity(j["id"], False))
        return True

    # Load the trap from file
    trap_fp = f"{base_dir}/{j['id']}_qc_traps.pickle"
    try:
        with open(trap_fp, "rb") as inp:
            trap_list = pickle.load(inp)
    except:
        print(f'Unable to load {trap_fp}, skipping job {j["id"]}')
        return False

    # Check trap validity
    try:
        trapper = BasicTrapper()
        validity = trapper.verify(trap_list, counts)
    except Exception as e:
        print("Got exception while verifying, contact dakk")
        print(e)
        return False

    # Send the set_result_validity
    if validity:
        trap_j = list(map(lambda t: t.dump(), trap_list))
        trap_file = f"{base_dir}/{j['id']}_traps.json"

        with open(trap_file, "w") as inp:
            inp.write(json.dumps(trap_j))

        trap_file_i = ipfs.upload(trap_file)

        print("\t", nb.set_result_validity(j["id"], True, trap_file_i))
    else:
        print("\t", nb.set_result_validity(j["id"], False))

    return True


def verifier_node():  # noqa: C901
    parser = default_parser()
    args = parser.parse_args()  # noqa: F841

    base_dir = create_dqpu_dirs()

    nb = NearBlockchain(args.account, args.network)
    ipfs = IPFSGateway(gateway="http://" + args.ipfs_gateway)  # noqa: F841

    # Start contract polling for new jobs
    running = True
    first_run = True
    n_vresult = 0
    n_verified = 0

    print("Verifier node started.")
    stats = repeat_until_done(lambda: nb.get_jobs_stats())

    while running:
        if first_run:
            first_run = False
            latest_jobs = repeat_until_done(lambda: nb.get_all_jobs())
        else:
            i = 0
            while (
                repeat_until_done(lambda: nb.get_jobs_stats())["validating-result"]
                == stats["validating-result"]
                and repeat_until_done(lambda: nb.get_jobs_stats())["pending-validation"]
                == stats["pending-validation"]
            ) and i < 5:
                time.sleep(random.randint(0, 5))
                i += 1

            latest_jobs = repeat_until_done(lambda: nb.get_latest_jobs())

        stats = repeat_until_done(lambda: nb.get_jobs_stats())
        random.shuffle(latest_jobs)

        # If there is a new job that needs validation, process it
        for j in latest_jobs:
            if j["status"] == "pending-validation":
                try:
                    if handle_pending_validation_job(j, ipfs, nb, base_dir):
                        n_verified += 1
                        stats["pending-validation"] -= 1
                except Exception as e:
                    print("\tFailed to handle pending-validation job:", e)
                    traceback.print_exc()

            elif (
                j["status"] == "validating-result"
                and j["verifier_id"] == nb.account.account_id
            ):
                try:
                    if handle_validating_result_job(j, ipfs, nb, base_dir):
                        n_vresult += 1
                        stats["validating-result"] -= 1
                except Exception as e:
                    print("\tFailed to handle validating-result job:", e)
                    traceback.print_exc()

        print(
            f"Account balance is {nb.balance():0.5f} N, job verified {n_verified}, "
            + f"result verified {n_vresult}"
        )



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qiskit import qasm2, transpile
from qiskit_aer import AerSimulator

from .sampler import Sampler


class AerSimulatorSampler(Sampler):
    def sample(self, shots):
        qc = qasm2.loads(
            self.circuit, custom_instructions=qasm2.LEGACY_CUSTOM_INSTRUCTIONS
        )  # .toQasmCircuit())
        simulator = AerSimulator()
        circ = transpile(qc, simulator)
        result = simulator.run(circ, shots=shots).result()
        counts = result.get_counts(circ)
        return counts



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import tempfile

from ..blockchain import repeat_until_done


def submit_job(nb, ipfs, qasm_data, num_qubits, depth, options):
    fp = tempfile.NamedTemporaryFile(mode="w", delete=False)
    fp.write(qasm_data)
    fp.close()

    job_file = ipfs.upload(fp.name)

    repeat_until_done(
        lambda: nb.submit_job(
            num_qubits, depth, options["shots"], job_file, options["reward"]
        )
    )
    return repeat_until_done(lambda: nb.get_latest_jobs())[-1]["id"]


def job_status(nb, jid):
    return repeat_until_done(lambda: nb.get_job_status(jid))


def job_remove(nb, jid):
    return repeat_until_done(lambda: nb.remove_job(jid))


def job_result(nb, ipfs, jid):
    j = repeat_until_done(lambda: nb.get_job(jid))
    data = json.loads(ipfs.get(j["result_file"]))
    trap_list = json.loads(ipfs.get(j["trap_file"]))
    return data, trap_list



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import random
import sys
import time
import traceback

from requests.exceptions import ReadTimeout

from .blockchain import (
    IPFSGateway,
    NearBlockchain,
    from_near,
    repeat_until_done,
    to_near,
)
from .cli import default_parser
from .sampler import SAMPLERS
from .utils import create_dqpu_dirs


def filter_jobs(jobs, args):
    filtered = []
    for j in jobs:
        if j["status"] != "waiting":
            continue

        # Check if reward/10 is < of max_deposit
        if int(j["reward_amount"]) / 10 > to_near(float(args.max_deposit)):
            print("reward / 10 is greater than max_deposit, skipping")
            continue

        # Check for max qubits
        if int(j["qubits"]) > int(args.max_qubits):
            print(
                f"qubits {j['qubits']} is greater than max_qubits {args.max_qubits}"
                + ", skipping"
            )
            continue

        # Check for min qubits
        if int(j["qubits"]) < int(args.min_qubits):
            print(
                f"qubits {j['qubits']} is lower than min_qubits {args.min_qubits}"
                + ", skipping"
            )
            continue

        filtered.append(j)

    return filtered


def handle_job(j, ipfs, nb, sampler_name, base_dir):
    # Get the qasm file
    try:
        jf = ipfs.get(j["job_file"], timeout=10)
    except ReadTimeout:  # TODO: move on ipfs.get raising a new exception
        print(f"\tTimeout getting file {j['job_file']}, skipping for now")
        return False

    print(f"\tGot file {j['job_file']}")

    # Load into a Sampler object (selected by params)
    # qc = Circuit.fromQasmCircuit(jf)
    sampler = SAMPLERS[sampler_name](jf)

    # Do the simulation
    print(f"\tStarting sampler {sampler_name}")
    t_start = time.time()
    counts = sampler.sample(j["shots"])
    t_duration = int(time.time() - t_start)
    if t_duration > 120:
        t_duration_s = (
            f"{int(t_duration / 60.)} minutes and {int(t_duration % 60)} seconds"
        )
    else:
        t_duration_s = f"{t_duration} seconds"

    # Upload the result
    result_f = f"{base_dir}/sampler/cache/{j['id']}_result.json"
    with open(result_f, "w") as cf:
        cf.write(json.dumps(counts))

    print(f"\tSampling done in {t_duration_s}, uploading {result_f}")
    jf_result = ipfs.upload(result_f)
    print(f"\tResult file uploaded {jf_result}")

    # Submit the result with the deposit
    try:
        sub_res = nb.submit_job_result(
            j["id"],
            jf_result,
            deposit=from_near(j["reward_amount"]) / 10 + 0.00001,
        )
        print(f"\t{sub_res}")
        return True
    except Exception as e:
        print("Failed to submit:", e)
        return False


def sampler_node():  # noqa: C901
    parser = default_parser()

    parser.add_argument("-d", "--max-deposit", help="maximum deposit", default=0.1)
    parser.add_argument(
        "-q", "--max-qubits", help="maximum number of simulable qubits", default=21
    )
    parser.add_argument("--min-qubits", help="minimum number of qubits", default=1)
    parser.add_argument(
        "-s",
        "--sampler",
        help="sampler to use",
        default="aersimulator",
        choices=SAMPLERS.keys(),
    )

    args = parser.parse_args()  # noqa: F841
    base_dir = create_dqpu_dirs()

    nb = NearBlockchain(args.account, args.network)
    ipfs = IPFSGateway(gateway="http://" + args.ipfs_gateway)  # noqa: F841

    # Start contract polling for new jobs
    running = True
    n_sampled = 0
    first_run = True

    print(f"Testing selected sampler: {args.sampler}")
    sampler_ok = SAMPLERS[args.sampler].test()
    if sampler_ok:
        print(f"Sampler {args.sampler} is working correctly.")
    else:
        print(f"Sampler {args.sampler} is not working correctly, exiting.")
        sys.exit()

    print("Sampler node started.")
    waiting_jobs = 0

    while running:
        if first_run:
            first_run = False
            latest_jobs = repeat_until_done(lambda: nb.get_all_jobs(True))
        else:
            i = 0
            while (
                repeat_until_done(lambda: nb.get_jobs_stats())["waiting"]
                == waiting_jobs
            ) and i < 5:
                time.sleep(random.randint(0, 5))
                i += 1
            latest_jobs = repeat_until_done(lambda: nb.get_latest_jobs())

        waiting_jobs = repeat_until_done(lambda: nb.get_jobs_stats()["waiting"])

        filtered_jobs = filter_jobs(latest_jobs, args)
        random.shuffle(filtered_jobs)
        fj_s = ", ".join(map(lambda j: j["id"], filtered_jobs))
        print(f"Found {len(filtered_jobs)} jobs matching sampler criteria: {fj_s}")
        i = 0

        # If there is a new job that needs execution, process it
        for j in filtered_jobs:
            i += 1
            print(
                f"[{i}/{len(filtered_jobs)}] Processing job {j['id']} with {j['qubits']} "
                + f"qubits for {j['shots']} shots"
            )
            try:
                if handle_job(j, ipfs, nb, args.sampler, base_dir):
                    n_sampled += 1
                    waiting_jobs -= 1
            except Exception as e:
                print("Failed to handle job:", e)
                traceback.print_exc()

        repeat_until_done(
            lambda: print(
                f"Account balance is {nb.balance():0.5f} N, sampled jobs {n_sampled}"
            )
        )



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


class Gate:
    def __init__(self, iden, matrix, nq=1):
        self.iden = iden
        self.nq = nq
        self._matrix = matrix

        try:
            self._tensor = np.reshape(self._matrix, (2, 2, 2, 2))
        except:
            self._tensor = None

    @property
    def matrix(self):
        if callable(self._matrix):
            raise Exception("This gate needs parameter")

        return self._matrix

    def parametrized(self, p):
        return Gate(self.iden, self._matrix(p), self.nq)


class GateCombination:
    def __init__(self, iden, gates, nq):
        self.iden = iden
        self.nq = nq
        self.gates = []  # Pairs of (Gate, qbidx)



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np

from .gate import Gate

# import cupy as np

c_one = complex(1, 0)
c_zero = complex(0, 0)
ket_zero = np.array([c_one, c_zero])
ket_one = np.array([c_zero, c_one])


# Standard gates


class Gates:
    P = Gate(
        "P",
        lambda k: np.array([[1, 0], [1, math.e ** (complex(0, 1) * k)]]),
    )
    R = Gate(
        "R",
        lambda k: np.array(
            [[1, 0], [1, math.e ** ((2 * math.pi * complex(0, 1)) / 2**k)]]
        ),
    )
    I = Gate("I", np.array([[1, 0], [0, 1]]))  # noqa: E741
    X = Gate("X", np.array([[0, 1], [1, 0]]))
    Y = Gate("Y", np.array([[0, -1.0j], [+1.0j, 0]]))
    Z = Gate("Z", np.array([[1, 0], [0, -1]]))
    S = Gate("S", np.array([[1, 0], [0, +1.0j]]))
    T = Gate("T", np.array([[1, 0], [0, math.e ** (+1.0j * math.pi / 4)]]))
    H = Gate("H", 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]]))
    SDG = Gate("SDG", np.array([[1, 0], [0, complex(0, -1)]]))

    CR = Gate(
        "CR", lambda k: np.kron(Gates.I.matrix, Gates.R.parametrized(k).matrix), 2
    )

    CX = Gate(
        "CX", np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]), 2
    )

    CY = Gate(
        "CY", np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1], [0, 0, +1.0j, 0]]), 2
    )

    CZ = Gate(
        "CZ", np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]]), 2
    )

    SWAP = Gate(
        "SWAP", np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]), 2
    )

    P0 = I  # Gate('P0', np.array([[1, 0], [0, 0]]))
    P1 = I  # Gate('P1', np.array([[0, 0], [0, 1]]))

    Gates1 = [
        I,
        X,
        # Y,
        Z,
        # S,
        # T,
        H,
        P,
        SDG,
    ]
    Gates2 = [CX, CY, CZ, SWAP]



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Sequence, Tuple

from ..q import Circuit, ExperimentResult


class TrapInfo:
    def __init__(self, trap_method, q_idx):
        self.trap_method = trap_method
        self.qubit = q_idx

    def dump(self, format="json"):
        if format != "json":
            raise Exception("Format not supported")

        return {"method": self.trap_method, "qubit": self.qubit}

    def loads(data, format="json"):
        raise Exception("Abstract")


class Trapper:
    """Trapper class"""

    def __init__(self):
        raise Exception("Abstract")

    def trap(
        self, qc: Circuit, level: Optional[int] = None
    ) -> Tuple[Circuit, Sequence[TrapInfo]]:
        """Add traps to the quantum circuits `qc`"""
        raise Exception("Abstract")

    def untrap(self, trapped_qc, traps: Sequence[TrapInfo]) -> Circuit:
        """Remove traps from the quantum circuits `trapped_qc`"""
        raise Exception("Abstract")

    def untrap_results(
        self, traps: Sequence[TrapInfo], results: ExperimentResult
    ) -> ExperimentResult:
        """Get the results for the original circuit, stripping away trap qubits"""
        qbits = list(map(lambda x: x.qubit, traps))
        qbits.sort(reverse=True)
        n_results: ExperimentResult = {}

        for bs, counts in results.items():
            n_bs = bs
            for q in qbits:
                n_bs = n_bs[::-1][:q] + n_bs[::-1][q + 1 :]
                n_bs = n_bs[::-1]

            if n_bs in n_results:
                n_results[n_bs] += counts
            else:
                n_results[n_bs] = counts

        return n_results

    def verify(self, traps: Sequence[TrapInfo], results: ExperimentResult) -> bool:
        raise Exception("Abstract")



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Sampler:
    """Abstract class that should be implemented by any sampler"""

    def __init__(self, circuit):
        if type(circuit) is not str:
            self.circuit = circuit.decode("ascii")
        else:
            self.circuit = circuit

    @classmethod
    def test(cls) -> bool:
        qc = (
            "OPENQASM 2.0;\n"
            'include "qelib1.inc";\n'
            "qreg q[2];\n"
            "creg c[2];\n"
            "h q[0];\n"
            "cx q[0], q[1];\n"
            "measure q -> c;"
        )
        c = cls(qc)
        r = c.sample(1024)

        if sorted(r.keys()) != ["00", "11"]:
            return False

        if (r["00"] + r["11"]) != 1024:
            return False

        return True

    def sample(self, shots):
        raise Exception("Not implemented")

    def compute(self):
        """Compute the statevector"""
        raise Exception("Not implemented")



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time


def repeat_until_done(f, n_iterations=10, wait_time=5):
    try:
        return f()
    except Exception as e:
        print(e)
        print(f"Function call failed, retrying in {wait_time} seconds ({n_iterations})")
        time.sleep(wait_time)
        return repeat_until_done(f, n_iterations - 1, wait_time)


class Blockchain:
    """Abstract blockchain provider"""

    def __init__(self, uri):
        pass

    def connect(self):
        raise Exception("Abstract: blockchain.connect")

    def contract_call(self, call_name, params):
        raise Exception("Abstract: blockchain.contract_call")

    def get_number_of_jobs(self) -> int:
        raise Exception("Abstract: blockchain.get_number_of_jobs")

    def get_jobs(self, from_index, limit):
        raise Exception("Abstract: blockchain.get_number_of_jobs")

    def get_jobs_paginated(self, from_index=0, to_index=None, limit=50, reverse=False):
        n = self.get_number_of_jobs() if to_index is None else to_index
        i = from_index
        js = []

        while i < n:
            js += self.get_jobs(from_index=i, limit=50)
            i += limit

        return js if not reverse else js[::-1]

    def get_all_jobs(self, reverse=False):
        return self.get_jobs_paginated(reverse=reverse)



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import subprocess

import requests


def start_ipfs_daemon():
    try:
        # Launch the IPFS daemon
        ipfs_process = subprocess.Popen(["ipfs", "daemon"])
        print("IPFS daemon started successfully.")
        return ipfs_process
    except FileNotFoundError:
        print(
            "IPFS executable not found. Please install IPFS first: "
            "https://docs.ipfs.tech/install/command-line/#install-official-binary-distributions"
        )
    except Exception as e:
        print(f"An error occurred while starting the IPFS daemon: {e}")


def stop_ipfs_daemon(ipfs_process):
    ipfs_process.terminate()


class IPFSGateway:
    def __init__(self, uri="http://localhost:5001", gateway="http://127.0.0.1:8080"):
        self.gateway = gateway
        self.uri = uri

    def upload(self, file_path):
        url = self.uri + "/api/v0/add"

        with open(file_path, "rb") as file:
            files = {"file": file}
            response = requests.post(url, files=files)

        if response.status_code == 200:
            ipfs_hash = response.text
            return json.loads(ipfs_hash)["Hash"]
        else:
            print(f"Error adding file to IPFS: {response.text}")

    def get(self, ipfs_hash, destination_file=None, timeout=10):
        # Set the URL for the IPFS daemon
        url = self.gateway + "/ipfs/"

        # Send a GET request to the IPFS daemon with the IPFS hash
        response = requests.get(f"{url}/{ipfs_hash}", timeout=timeout)

        # Check if the request was successful
        if response.status_code == 200:
            # Save the file content
            file_content = response.content

            if destination_file:
                with open(destination_file, "wb") as file:
                    file.write(file_content)
            return file_content
        else:
            print(f"Error retrieving file from IPFS: {response.text}")



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import random
from typing import Optional, Sequence, Tuple, cast

from ..q import Circuit, ExperimentResult, Gates
from .trapper import TrapInfo, Trapper


# TODO: make it base for other strategies, or move helpers to Trapper
class BasicTrapInfo(TrapInfo):
    def __init__(self, q_idx, val, prob=1.0):
        super().__init__("basic", q_idx)
        self.value_expected = val
        self.probability = prob

    def __repr__(self):
        return (
            f"qubit {self.qubit} expect to have value {self.value_expected} "
            f"with probability {self.probability} trapped with {self.trap_method}"
        )

    def dump(self, format="json"):
        if format != "json":
            raise Exception("Format not supported")

        data = super().dump(format)
        data["value_expected"] = self.value_expected
        data["probability"] = self.probability
        return data

    @staticmethod
    def loads(data, format="json"):
        if format != "json":
            raise Exception("Format not supported")

        return BasicTrapInfo(data["qubit"], data["value_expected"], data["probability"])


class BasicTrapper(Trapper):
    """A basic trapper, just adds `n` qubits initialized in a random choice of |0> and |1>"""

    def __init__(self):
        pass

    def trap(
        self, qc: Circuit, level: Optional[int] = None
    ) -> Tuple[Circuit, Sequence[BasicTrapInfo]]:
        """Add traps to the quantum circuit"""
        if level is None:
            level = 1

        qc = copy.deepcopy(qc)

        traps: Sequence[BasicTrapInfo] = []
        gates = qc.gates
        for i in range(level):
            qc.n_qbits += 1
            i_r = random.randint(0, qc.n_qbits - 1)

            def remap_qbit(q_idx):
                return q_idx + 1 if q_idx >= i_r else q_idx

            def remap_gate(gq):
                a, p, args = gq
                p = list(map(remap_qbit, p))
                return (a, p, args)

            def remap_trap(t):
                t.qubit = remap_qbit(t.qubit)
                return t

            gates = list(map(remap_gate, gates))
            traps = list(map(remap_trap, traps))

            v_e = False
            if not random.choice([True, False]):
                v_e = True
                gates.insert(random.randint(0, len(gates) - 1), (Gates.X, [i_r]))

            traps.append(BasicTrapInfo(i_r, v_e))

        qc.gates = gates
        return (qc, traps)

    def untrap(self, trapped_qc, traps: Sequence[TrapInfo]) -> Circuit:
        """Remove traps from the quantum circuits `trapped_qc`"""
        raise Exception("TODO")

    def verify(self, traps: Sequence[TrapInfo], results: ExperimentResult) -> bool:
        """Get bitstring result for trap qubits, and check for the result"""
        tl = {}

        for t in traps:
            tl[t.qubit] = {"0": 0, "1": 0}

        for bs, counts in results.items():
            bs = bs[::-1]  # This is for qiskit
            for t in traps:
                idx = t.qubit

                if bool(int(bs[idx])):
                    tl[idx]["1"] += counts
                else:
                    tl[idx]["0"] += counts

        for t in traps:
            v = tl[t.qubit]
            v_prob = abs((v["1"] - v["0"]) / (v["1"] + v["0"]))

            t_ = cast(BasicTrapInfo, t)
            if v_prob < t_.probability - 0.05 and v_prob > t_.probability + 0.05:
                return False

            vmax = False if v["0"] > v["1"] else True
            if vmax != t_.value_expected:
                return False

        return True



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def qiskit_execute(c):
    from qiskit import transpile
    from qiskit_aer import AerSimulator

    qc = c.toQiskitCircuit()
    qc.measure_all()

    sim = AerSimulator(method="statevector")
    qc.save_statevector()
    circ = transpile(qc, sim)
    job = sim.run(circ, shots=1024)
    result = job.result()
    counts = result.get_counts()
    outputstate = result.get_statevector(circ, decimals=3)

    # i = 0
    # for x in outputstate:
    #     print (f'{fixbinlen(bin(i)[2:], c.n)} => {np.absolute(x) ** 2}')
    #     i+=1
    return outputstate, counts


def fixbinlen(v, nq=3):
    return ("0" * (nq - len(v)) + v)[::-1]


def states_string_rep(nq=3):
    state_strs = [bin(x) for x in range(2**nq)]
    for i in range(len(state_strs)):
        state_strs[i] = fixbinlen(state_strs[i][2:], nq)

    # state_strs = ['' for _ in range(2**nq)]
    # basis_strs = ['0', '1']

    # for q in range(nq):
    # 	for i in range(len(state_strs)):
    # 		b = basis_strs[((i//(2**q))) % 2]
    # 		state_strs[i] =  state_strs[i] + b

    return state_strs


def basis_state_probs(svec):
    return np.array([np.absolute(s) ** 2 for s in svec])


def print_output(nq, out):
    xx = basis_state_probs(out)
    yy = states_string_rep(nq)

    for i in range(2**nq):
        print(yy[i], xx[i])


def plot_output(nq, out):
    import matplotlib.pyplot as plt

    nq = int(np.log2(out.shape[0]))
    fig, ax = plt.subplots(1, 1)
    ax.bar(range(2**nq), basis_state_probs(out))
    ax.set_xticks(range(2**nq))
    ax.set_xticklabels(states_string_rep(nq))
    ax.set_ylim(-0, 1)
    ax.grid(True)
    ax.set_ylabel(r"$P(S_c)$")
    ax.set_xlabel(r"$S_c$")
    plt.show()



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import openqasm3
from openqasm3 import ast as oast
from openqasm3 import printer

from .gate import Gate
from .gates import Gates


# Abstract class representing a circuit; it doesn't hold a quantum state
class Measure:
    def __init__(self):
        self.iden = "measure"


class Barrier:
    def __init__(self):
        pass


class Circuit:
    def __init__(self, qbn, cbn):
        self.n_qbits = qbn
        self.n_cbits = cbn
        self.gates = []

    def __iter__(self):
        yield from self.gates

    @staticmethod
    def random(n, dp, measure=False):
        qc = Circuit(n, n)

        for y in range(dp - 1):
            for x in range(n):
                if random.choice([True, False]) or n == 1:
                    qc.apply(
                        random.choice([Gates.X, Gates.H, Gates.Y, Gates.Z, Gates.T]),
                        [x],
                    )
                else:
                    r1 = random.randint(0, n - 1)
                    r2 = random.randint(0, n - 1)

                    while r1 == r2:
                        r1 = random.randint(0, n - 1)
                        r2 = random.randint(0, n - 1)

                    qc.apply(
                        random.choice([Gates.CX, Gates.CZ]), (r1, r2)
                    )  # Gates.SWAP,

        if measure:
            for x in range(n):
                qc.measure(x)

        return qc

    def apply(self, g, qb):
        self.gates.append((g, qb))

    def measure(self, qb, cb=None):
        if cb is None:
            cb = qb
        self.gates.append((Measure(), (qb, cb)))

    def draw(self):
        print(f"QC({self.n_qbits},{self.n_cbits})", end=" ==> ")
        for x in self.gates:
            if isinstance(x[0], Measure):
                print(f"Measure({x[1]})", end=" -- ")
            elif isinstance(x[0], Gate):
                print(f"{x[0].iden}({x[1]})", end=" -- ")
        print()

    def buildGraph(self):
        import matplotlib.pyplot as plt
        import networkx as nx

        G = nx.Graph()

        for x in self.gates:
            a, p = x

            if isinstance(a, Gate) and a.nq > 1:
                G.add_edge(p[0], p[1])

        options = {
            "font_size": 12,
            "node_size": 300,
            # "node_color": "white",
            # "edgecolors": "black",
            "linewidths": 1,
            "width": 1,
        }
        nx.draw_networkx(G, **options)

        # Set margins for the axes so that nodes aren't clipped
        ax = plt.gca()
        ax.margins(0.0010)
        plt.axis("off")
        plt.show()

    @staticmethod
    def fromQasmCircuit(qasm_data: str):  # noqa: C901
        def qubit_to_i(q):
            if not isinstance(q, oast.IndexedIdentifier):
                raise Exception("only indexed identifier allowed")
            if len(q.indices) != 1:
                raise Exception("only indexed identifier with 1 elemtn allowed")
            return q.indices[0][0].value

        qa = openqasm3.parser.parse(qasm_data)

        n_q = None
        n_c = None
        gates = []
        end = False

        for s in qa.statements:
            if end:
                raise Exception("Last statement should me a measurement")

            if isinstance(s, oast.QubitDeclaration):
                if n_q is not None:
                    raise Exception("Only one qubit register is allowed")
                n_q = s.size.value
            elif isinstance(s, oast.ClassicalDeclaration):
                if n_c is not None:
                    raise Exception("Only one classical register is allowed")
                n_c = s.type.size.value
            elif isinstance(s, oast.QuantumGate):
                args = list(map(printer.dumps, s.arguments))

                p = list(map(qubit_to_i, s.qubits))
                gn = s.name.name.upper()
                if not hasattr(Gates, gn):
                    raise Exception(f"Unknown gate {gn}")

                g = getattr(Gates, gn)
                gates.append((g, p, args))
            elif isinstance(s, oast.QuantumMeasurementStatement):
                end = True
            elif isinstance(s, oast.Include):
                pass
            else:
                raise Exception(f"Unhandled {s}")

        qc = Circuit(n_q, n_c)
        qc.gates = gates
        return qc

    def toQasmCircuit(self):
        qasm = "OPENQASM 2.0;\n"
        qasm += 'include "qelib1.inc";\n'
        qasm += "qreg q[" + str(self.n_qbits) + "];\n"
        qasm += "creg c[" + str(self.n_qbits) + "];\n"

        for x in self.gates:
            args = ""

            if len(x) == 2:
                x = (x[0], x[1], [])

            a, p, ar = x
            if len(ar) > 0:
                args = ("(" + (",".join(ar)) + ")").replace(" ", "")

            qbs = ", ".join(map(lambda g: "q[" + str(g) + "]", p))
            qasm += f"{a.iden.lower()}{args} {qbs};\n"

        qasm += "measure q -> c;"

        return qasm

    def toQiskitCircuit(self):  # noqa: C901
        from qiskit import QuantumCircuit

        qc = QuantumCircuit(self.n_qbits, 0)

        for x in self.gates:
            a, p = x

            if isinstance(a, Gate):
                if a.iden == "CX":
                    qc.cx(p[0], p[1])
                elif a.iden == "CZ":
                    qc.cz(p[0], p[1])
                elif a.iden == "CY":
                    qc.cy(p[0], p[1])
                elif a.iden == "SWAP":
                    qc.swap(p[0], p[1])
                elif a.iden == "X":
                    qc.x(p)
                elif a.iden == "I":
                    qc.i(p)
                elif a.iden == "Z":
                    qc.z(p)
                elif a.iden == "Y":
                    qc.y(p)
                elif a.iden == "H":
                    qc.h(p)
                elif a.iden == "S":
                    qc.s(p)
                elif a.iden == "T":
                    qc.t(p)

            elif isinstance(a, Measure):
                qc.measure(p, p)

        return qc



# Copyright 2024 Davide Gessa

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

import dqpu

REQUIREMENTS = [
    # "dask[distributed]==2024.4.2",
    "scipy==1.13.0",
    "matplotlib",
    "qiskit==1.0.2",
    "qiskit_aer==0.14.1",
    "py-near",
    "requests",
    "numpy==1.26.4",
    "openqasm3[parser]",
    "scikit-build",
    "base58",
    "loguru",
    "pydantic",
    "nest_asyncio",
    "pyqrack",
    "qiskit-qrack-provider>=0.11.0",
]

setup(
    name="dqpu",
    version=dqpu.__version__,
    python_requires=">= 3.9.2,<3.12",
    description="",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Davide Gessa",
    setup_requires="setuptools",
    author_email="gessadavide@gmail.com",
    license="Apache 2.0",
    packages=[
        "dqpu",
        "dqpu.q",
        "dqpu.sampler",
        "dqpu.verifier",
        "dqpu.blockchain",
        "dqpu.backends",
        "dqpu.backends.qiskit",
    ],
    zip_safe=False,
    install_requires=REQUIREMENTS,
    extras_require={},
    entry_points={
        "console_scripts": [
            "dqpu-sim_trap_test_main = dqpu.sim_trap_test_main:main",
            "dqpu-verifier = dqpu.verifiernode:verifier_node",
            "dqpu-sampler = dqpu.samplernode:sampler_node",
            "dqpu-cli = dqpu.cli:cli",
        ],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Physics",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/dakk/dqpu",
    project_urls={
        "Bug Tracker": "https://github.com/dakk/dqpu/issues/",
        "Documentation": "https://dakk.github.io/dqpu",
        "Source": "https://github.com/dakk/dqpu",
    },
)
